#include "userprog/syscall.h"
#include <stdio.h>
#include <syscall-nr.h>
#include "threads/interrupt.h"
#include "threads/thread.h"

#include "file_descript.h"

static void syscall_handler (struct intr_frame *);
static void sys_halt (void);
static void sys_exit (int);
static int sys_exec (const char *);
static int sys_wait (int);
static bool sys_create (const char *, unsigned);
static bool sys_remove (const char *);
static int sys_open (const char *);
static bool sys_filesize (int);
static int sys_read (int, void *, unsigned);
static int sys_write (int, const void *, unsigned);
static void sys_seek (int, unsigned);
static unsigned sys_tell (int);
static void sys_close (int);

void
syscall_init (void)
{
  intr_register_int (0x30, 3, INTR_ON, syscall_handler, "syscall");

}

static void sys_halt (void)
{
  asm volatile ("hlt");
}

static void sys_exit (int status)
{
  printf ("Exit code : %d\n", status);
  thread_exit ();
}

static int sys_exec (const char *cmd_line)
{
  return 0;
}

static int sys_wait (int pid)
{
  return 0;
}

static bool sys_create (const char *file, unsigned initial_size)
{
  return false;
}

static bool sys_remove (const char *file)
{
  return false;
}

static int sys_open (const char *file)
{
  return false;
}

static bool sys_filesize (int fd)
{
  return false;
}

static int sys_read (int fd, void *buffer, unsigned size)
{
  return false;
}

static int sys_write (int fd, const void *buffer, unsigned size)
{
  struct fd *fd_data = fd_fdnum_to_fd (fd);
  int i 
  return false;
}

static void sys_seek (int fd, unsigned position)
{
}

static unsigned sys_tell (int fd)
{
  return false;
}

static void sys_close (int fd)
{
}

static void
syscall_handler (struct intr_frame *f UNUSED) 
{
  int syscall_no = *(int *)(f->esp);
  
  switch (syscall_no)
    {
    case SYS_HALT:
    {
      sys_halt ();
    }
    break;

    case SYS_EXIT:
    {
      int pid = *(int *)(f->esp + 4);

      sys_exit (pid);
    }
    break;

    case SYS_EXEC:
    {
      const char *cmd_line = *(char **)(f->esp + 4);

      f->eax = sys_exec (cmd_line);
    }
    break;

    case SYS_WAIT:
    {
      int pid = *(int *)(f->esp + 4);

      f->eax = sys_wait (pid);
    }
    break;

    case SYS_CREATE:
    {
      const char *file = *(char **)(f->esp + 4);
      unsigned initial_size = *(int *)(f->esp + 8);

      f->eax = sys_create (file, initial_size);
    }
    break;

    case SYS_REMOVE:
    {
      const char *file = *(char **)(f->esp + 4);

      f->eax = sys_remove (file);
    }
    break;

    case SYS_OPEN:
    {
      const char *file = *(char **)(f->esp + 4);

      f->eax = sys_open (file);
    }
    break;

    case SYS_FILESIZE:
    {
      int fd = *(int *)(f->esp + 4);

      f->eax = sys_filesize (fd);
    }
    break;

    case SYS_READ:
    {
      int fd = *(int *)(f->esp + 4);
      const void *buffer = *(void **)(f->esp + 8);
      unsigned size = *(unsigned *)(f->esp + 12);

      f->eax = sys_read (fd, buffer, size);
    }
    break;

    case SYS_WRITE:
    {
      int fd = *(int *)(f->esp + 4);
      const void *buffer = *(void **)(f->esp + 8);
      unsigned size = *(unsigned *)(f->esp + 12);

      f->eax = sys_write (fd, buffer, size);
    }
    break;

    case SYS_SEEK:
    {
      int fd = *(int *)(f->esp + 4);
      unsigned position = *(unsigned *)(f->esp + 8);

      sys_seek (fd, position);
    }
    break;

    case SYS_TELL:
    {
      int fd = *(int *)(f->esp + 4);

      f->eax = sys_tell (fd);
    }
    break;

    case SYS_CLOSE:
    {
      int fd = *(int *)(f->esp + 4);

      sys_close (fd);
    }
    break;
    default:
    break;
    }
}


/* File Descriptor maker.  */


static inline bool fd_isnull (int);

void
fd_init ()
{
  memset (fd_bitmap, 0, sizeof (int) * 32);
  memset (fd_table, 0, sizeof (struct fd *) * 128);

  struct fd *stdin;
  struct fd *stdout;
  struct fd *stderr;

  fd_reserve_bitmap (0);
  fd_reserve_bitmap (1);
  fd_reserve_bitmap (2);

  stdin = palloc_get_page (1);
  stdout = palloc_get_page (1);
  stderr = palloc_get_page (1);

  stdin->fd_dev_id = 0;
  stdout->fd_dev_id = 1;
  stderr->fd_dev_id = 2;

  stdin->dev.dev_id = 0;
  stdout->dev.dev_id = 1;
  stderr->dev.dev_id = 2;

  strlcpy (stdin->dev.dev_name, "stdin", 6);
  strlcpy (stdout->dev.dev_name, "stdout", 7);
  strlcpy (stderr->dev.dev_name, "stderr", 7);

  fd_table[0] = stdin;
  fd_table[1] = stdout;
  fd_table[2] = stderr;
}

static inline bool
fd_isnull (int num)
{

  return !(fd_bitmap[num / 32] & (1 << (num % 32)));
}

void
fd_reserve_bitmap (int fdnum)
{
  fd_bitmap[fdnum / 32] |= 1 << (fdnum % 32);
}

bool
fd_find_fdnum ()
{
  int i;
  for (i = 0; i < 128; i++)
    {
      if (fd_isnull (i))
        return i;
    }
  return -1;
}

struct
fd *fd_fdnum_to_fd (int num)
{
  return fd_table[num];
}

int
fd_fd_to_fdnum (struct fd *fd)
{
  return fd->fd_dev_id;
}
